# -*- coding: utf-8 -*-

# Copyright 2008-2012 Richard Dymond (rjdymond@gmail.com)
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.

from .skoolhtml import Table, Cell, join
from .microsphere import MicrosphereHtmlWriter, Udg, MicrosphereAsmWriter

class BackToSkoolHtmlWriter(MicrosphereHtmlWriter):
    def init(self):
        MicrosphereHtmlWriter.init(self)
        self.init_lessons(37, 59)
        self.init_characters(183, 209)
        self.font_address = 55040
        self._calculate_tap_index(59392, 88)

    def _get_paths(self):
        paths = MicrosphereHtmlWriter._get_paths(self)
        paths['MutableImagePath'] = join('images', 'mutables')
        return paths

    def get_skool_udg(self, y, x, show_chars=False):
        q_addr = x + 46336
        q = self.snapshot[q_addr]
        h = y + 160
        hl = 180 + q // 2 + 256 * h
        i = 0
        if q & 1:
            p = self.snapshot[hl] & 15
        else:
            p = (self.snapshot[hl] & 240) // 16
        if p == 0:
            i = q & 3 or 2
            if q & 1:
                p = self.snapshot[hl + 256] & 15
            else:
                p = (self.snapshot[hl + 256] & 240) // 16
        attr = i | (8 * p)
        ref_addr = q + 256 * h
        ref = self.snapshot[ref_addr]
        a = 136 // 2**(q & 3) & self.snapshot[144 + q // 4 + 256 * h]
        udg_page = (8 if a & 15 else 0) + (144 if a & 240 else 128)
        udg_addr = ref + 256 * udg_page
        udg_bytes = self.snapshot[udg_addr:udg_addr + 2048:256]
        skool_udg = Udg(attr, udg_bytes, q_addr=q_addr, ref_addr=ref_addr, ref=ref, udg_page=udg_page, x=x, y=y)
        return skool_udg

    def get_sprite_udg(self, state, attr, row, col, udg_page=None):
        as_norm = state | 128
        udg_page = 183 if as_norm < 208 else 199
        col = 2 - col if state & 128 else col
        ref_page = 215 + col * 4 + row
        udg = self._get_sprite_udg(as_norm, attr, ref_page, udg_page)
        if state & 128:
            udg.flip()
        return udg

    def alter_skool_udgs(self, address):
        coords = []
        while True:
            y = self.snapshot[address]
            if y == 255:
                break
            x = self.snapshot[address + 1]
            coords.append((x, y))
            udg_lsb = self.snapshot[address + 2]
            attr = self.snapshot[address + 3]
            d = y + 160
            col_ref = self.snapshot[x + 46336]
            self.snapshot[col_ref + 256 * d] = udg_lsb
            z = 2**(col_ref & 3)
            e = 128 + (col_ref + 104) // 2
            a = self.snapshot[e + 256 * d]
            if col_ref & 1:
                a = (a & 240) | (attr & 15)
            else:
                a = (16 * (attr & 15)) | (a & 15)
            self.snapshot[e + 256 * d] = a
            a = attr & 192
            if a & 64:
                a = a - 56
            e = 144 + col_ref // 4
            self.snapshot[e + 256 * d] = self.snapshot[e + 256 * d] & 255 - 136 // z | a // z
            address += 4
        return coords

    def _get_skool_udgs_by_coords(self, coords):
        left_x, top_y, width, height = self._get_dimensions(coords)
        udg_array = []
        for y in range(top_y, top_y + height):
            udg_array.append([None] * width)
        for x, y in coords:
            udg_array[y - top_y][x - left_x] = self.get_skool_udg(y, x)
        return udg_array

    def _get_dimensions(self, coords, padding=0):
        x_range = [c[0] for c in coords]
        y_range = [c[1] for c in coords]
        left_x = min(x_range)
        top_y = min(y_range)
        width = max(x_range) - left_x + 1
        height = max(y_range) - top_y + 1
        left_x -= padding
        top_y -= padding
        width += 2 * padding if left_x + width < 192 else padding
        height += 2 * padding if top_y + height < 20 else padding
        return left_x, top_y, width, height

    def get_mutable_udg_array(self, address, padding):
        self.push_snapshot()
        coords = self.alter_skool_udgs(address)
        if padding >= 0:
            left_x, top_y, width, height = self._get_dimensions(coords, padding)
            udg_array = self.get_skool_udgs(left_x, top_y, width, height)
        else:
            udg_array = self._get_skool_udgs_by_coords(coords)
        self.pop_snapshot()
        return udg_array

    def get_initial_states(self):
        return [self.snapshot[32 + 256 * c:35 + 256 * c] for c in range(183, 211)]

    def write_mutable_tile_tables(self, cwd):
        lines = []

        doors = []
        doors.append(56893) # Left study door
        doors.append(57344) # Right study door
        doors.append(56832) # Science Lab storeroom door
        doors.append(57405) # Boys' skool door
        doors.append(64000) # Skool gate
        doors.append(56320) # Drinks cabinet door

        windows = []
        windows.append(55296) # Top floor window
        windows.append(55808) # Middle floor window

        bike = [57856] # Tree with bike attached

        sections = (('Doors (open)', doors), ('Windows (open)', windows), ('Tree (with bike attached)', bike))
        for title, section in sections:
            lines.append('<h3>%s</h3>\n<table class="mgrid">\n<tr>\n' % title)
            for addr in section:
                mutable_udgs = self.get_mutable_udg_array(addr, -1)
                lines.append('<td>\n')
                lines += self.write_playarea_tile_table(cwd, mutable_udgs, 'm')
                lines.append('</td>\n')
            lines.append('</tr>\n</table>\n')

        cups = []
        cups.append((56694, "Leftmost cup on the shelf in the boys' skool (water/sherry)"))
        cups.append((56950, "Middle cup on the shelf in the boys' skool (water/sherry)"))
        cups.append((57206, "Rightmost cup on the shelf in the boys' skool (water/sherry)"))
        cups.append((57462, "Cup on the shelf in the girls' skool (water/sherry)"))
        lines.append('<h3>Cups</h3>\n<table>\n')
        for a, desc, in cups:
            lines.append('<tr>\n<td>\n<table class="mgrid">\n<tr>\n')
            for address in (a, a + 5):
                cup_udg_array = self.get_mutable_udg_array(address, -1)
                lines.append('<td>\n')
                lines += self.write_playarea_tile_table(cwd, cup_udg_array, 'c%i_' % address)
                lines.append('</td>\n')
            lines.append('</tr>\n</table>\n</td><td>%s</td></tr>\n' % desc)
        lines.append('</table>\n')

        return lines

    def keypress_table(self, cwd):
        keypress_routines = self.keypress_routines

        table = Table('keytable', 'Index', 'Key', 'Address', 'Offset', 'Lookup', 'Routine', 'Purpose')
        for index in range(80):
            key = self.get_chr(index + 48)
            address = index + 58624
            offset = self.snapshot[address]
            lookup = routine_link = purpose = ''
            if offset:
                lookup = offset + 58624
                routine = self.snapshot[lookup] + 256 * self.snapshot[lookup + 1]
                purpose = keypress_routines[routine]
                routine_link = self.expand('#R%i' % routine, cwd)
            row = []
            row.append(Cell(index, "keyIndex"))
            row.append(Cell(key, "key"))
            row.append(Cell(address, "keyAddress"))
            row.append(Cell(offset, "keyOffset"))
            row.append(Cell(lookup, "keyAddress"))
            row.append(Cell(routine_link, "keyAddress"))
            row.append(Cell(purpose, "keyPurpose"))
            table.add_row(row)

        return table.html()

    def get_playarea_tile_text(self, tile):
        text = ['<i>Q_addr</i>=(x,181)=%i, <i>Q</i>=%i' % (tile.q_addr, self.snapshot[tile.q_addr])]
        text.append('<i>ref_addr</i>=(<i>Q</i>,y+160)=%i' % tile.ref_addr)
        text.append('<i>ref</i>=%i, <i>udg_page</i>=%i' % (tile.ref, tile.udg_page))
        text.append('<i>udg_addr</i>=(<i>ref</i>,<i>udg_page</i>)=%i' % (tile.ref + 256 * tile.udg_page))
        text.append('<i>attr</i>=%i' % tile.attr)
        return text

    def mutable(self, cwd, address, padding=0):
        fname = 'mutable%i_%i' % (address, padding) if padding else 'mutable%i' % address
        mutable_path = self.image_path(fname, 'MutableImagePath')
        if self.need_image(mutable_path):
            mutable_udgs = self.get_mutable_udg_array(address, padding)
            self.write_image(mutable_path, mutable_udgs, scale=2)
        return self.img_element(cwd, mutable_path)

    def adjust_mutable(self, cwd, address):
        self.alter_skool_udgs(address)

    def place_char(self, cwd, char_num, x=None, y=None, state=None):
        base_addr = 32 + 256 * char_num
        if state is not None:
            self.snapshot[base_addr] = state
        if x is not None:
            self.snapshot[base_addr + 1] = x
        if y is not None:
            self.snapshot[base_addr + 2] = y

    def hide_chars(self, cwd=None):
        for char_num in range(183, 215):
            self.place_char(cwd, char_num, 200)

class BackToSkoolAsmWriter(MicrosphereAsmWriter):
    pass
